package org.processmining.dataawarecnetminer.converter;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import org.processmining.datapetrinets.DataPetriNet;
import org.processmining.datapetrinets.DataPetriNet.PetrinetWithMarkings;
import org.processmining.datapetrinets.DataPetriNetsWithMarkings;
import org.processmining.datapetrinets.exception.NonExistingVariableException;
import org.processmining.datapetrinets.expression.GuardExpression;
import org.processmining.models.cnet.CNetBinding;
import org.processmining.models.cnet.CNetNode;
import org.processmining.models.cnet.CausalNet;
import org.processmining.models.dcnet.DataAwareCausalNet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetGraph;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.DataElement;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PNWDTransition;
import org.processmining.models.semantics.petrinet.Marking;

import com.google.common.collect.BiMap;
import com.google.common.collect.FluentIterable;
import com.google.common.collect.HashBiMap;
import com.google.common.collect.ImmutableSet;

public class CausalNetToPetrinet {

	private static final class Gateway {

		private final CNetNode source;
		private final CNetNode target;

		Gateway(CNetNode source, CNetNode target) {
			this.source = source;
			this.target = target;
		}

		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + ((source == null) ? 0 : source.hashCode());
			result = prime * result + ((target == null) ? 0 : target.hashCode());
			return result;
		}

		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (getClass() != obj.getClass())
				return false;
			Gateway other = (Gateway) obj;
			if (source == null) {
				if (other.source != null)
					return false;
			} else if (!source.equals(other.source))
				return false;
			if (target == null) {
				if (other.target != null)
					return false;
			} else if (!target.equals(other.target))
				return false;
			return true;
		}

		@Override
		public String toString() {
			return "Gateway from " + source + " to " + target;
		}

	}

	private static final class LabelGeneratorImpl implements LabelGenerator {

		private String prefix;
		private int index;

		LabelGeneratorImpl(String prefix) {
			this.prefix = prefix;
		}

		public String newLabel() {
			return prefix.concat(String.valueOf((index++)));
		}
	}

	private interface LabelGenerator {
		String newLabel();
	}

	private LabelGenerator placeLabelGenerator;
	private LabelGenerator transitionLabelGenerator;
	private boolean isReduceNet = true;
	private BiMap<CNetBinding, Transition> bindingToTransition;
	private BiMap<CNetNode, Transition> nodeToTransition;

	public PetrinetWithMarkings convert(final CausalNet causalNet) {
		final DataPetriNetsWithMarkings dpn = convert(DataAwareCausalNet.Factory.fromCausalNet(causalNet));
		return DataPetriNet.Factory.toPetrinetWithMarkings(dpn);
	}

	public DataPetriNetsWithMarkings convert(DataAwareCausalNet dataCausalNet) {

		placeLabelGenerator = new LabelGeneratorImpl("p");
		transitionLabelGenerator = new LabelGeneratorImpl("t");
		bindingToTransition = HashBiMap.create();
		nodeToTransition = HashBiMap.create();

		final CausalNet causalNet = dataCausalNet.getCausalNet();
		final DataPetriNetsWithMarkings net = DataPetriNet.Factory.create("");
		final Marking initialMarking = new Marking();
		net.setInitialMarking(initialMarking);
		final Marking finalMarking = new Marking();
		net.setFinalMarkings(new Marking[] { finalMarking });

		Map<CNetNode, Transition> transitions = new HashMap<>();
		Map<CNetNode, Place> inputPlaces = new HashMap<>();
		Map<CNetNode, Place> outputPlaces = new HashMap<>();
		Map<Gateway, Place> gatewayPlaces = new HashMap<>();

		for (CNetNode currentTask : causalNet.getNodes()) {
			if (hasRelations(currentTask, causalNet)) {
				if (currentTask.equals(causalNet.getStartNode())) {
					Place source = net.addPlace("source");
					initialMarking.add(source);
					outputPlaces.put(currentTask, source);
				} else if (currentTask.equals(causalNet.getEndNode())) {
					Place sink = net.addPlace("sink");
					finalMarking.add(sink);
					inputPlaces.put(currentTask, sink);
				} else {
					Transition currentTransition = addTransitionIfNeeded(net, currentTask, transitions);
					nodeToTransition.put(currentTask, currentTransition);
					for (String write : dataCausalNet.getWrites(currentTask)) {
						net.assignWriteOperation(currentTransition, addVariableIfNeeded(net, dataCausalNet, write));
					}
					Place inputPlace = addPlaceIfNeeded(currentTask, inputPlaces, net, placeLabelGenerator);
					Place outputPlace = addPlaceIfNeeded(currentTask, outputPlaces, net, placeLabelGenerator);
					net.addArc(inputPlace, currentTransition);
					net.addArc(currentTransition, outputPlace);
				}
			}
		}

		for (CNetNode currentTask : causalNet.getNodes()) {

			Place outPlace = outputPlaces.get(currentTask);
			Set<CNetBinding> outputBindings = causalNet.getOutputBindings(currentTask);

			if (outPlace != null && outputBindings.isEmpty()) {
				finalMarking.add(outPlace);
			} else if (outPlace == null && !outputBindings.isEmpty()) {
				throw new IllegalArgumentException(
						"Invalid C-Net structure: Artifical end transition should not have any output binding, but is assigned output bindings "
								+ outputBindings);
			}

			for (CNetBinding outputBinding : outputBindings) {

				PNWDTransition split = addAnonymousTransition(net, transitionLabelGenerator);

				bindingToTransition.put(outputBinding, split);

				if (dataCausalNet.getGuardedBindings().contains(outputBinding)) {
					GuardExpression guardExpression = dataCausalNet.getGuardExpression(outputBinding);
					addVariablesIfNeeded(net, guardExpression.getNormalVariables(), dataCausalNet);
					try {
						split.setGuard(net, guardExpression);
					} catch (NonExistingVariableException e) {
						throw new RuntimeException(e);
					}
				}

				net.addArc(outPlace, split);

				for (CNetNode nextTask : outputBinding.getBoundNodes()) {
					Place gateway = addGatewayIfNeeded(currentTask, nextTask, gatewayPlaces, net, placeLabelGenerator);
					net.addArc(split, gateway);
				}
			}
		}

		for (CNetNode currentTask : causalNet.getNodes()) {

			Place inPlace = inputPlaces.get(currentTask);
			Set<CNetBinding> inputBindings = causalNet.getInputBindings(currentTask);

			if (inPlace != null && inputBindings.isEmpty()) {
				initialMarking.add(inPlace);
			}

			for (CNetBinding inputBinding : inputBindings) {

				Transition join = addAnonymousTransition(net, transitionLabelGenerator);
				bindingToTransition.put(inputBinding, join);

				net.addArc(join, inPlace);

				for (CNetNode inputTask : inputBinding.getBoundNodes()) {
					Place gateway = addGatewayIfNeeded(inputTask, currentTask, gatewayPlaces, net, placeLabelGenerator);
					net.addArc(gateway, join);
				}
			}
		}

		if (isReduceNet) {
			// remove irrelevant invisible transitions
			for (Transition t : ImmutableSet.copyOf(net.getTransitions())) {
				if (t.isInvisible()) {
					if (isSimpleSequenceFlow(t)) {
						removeFromNet(net, t, initialMarking, finalMarking);
					}
				}
			}
			for (Place p : ImmutableSet.copyOf(net.getPlaces())) {
				if (isSimpleSequenceFlow(p) && isTauPlace(p)) {
					removeFromNet(net, p, initialMarking, finalMarking);
				}
			}
		}

		return net;
	}

	private void addVariablesIfNeeded(DataPetriNet net, Set<String> normalVariables, DataAwareCausalNet dataCausalNet) {
		for (String variable : normalVariables) {
			addVariableIfNeeded(net, dataCausalNet, variable);
		}
	}

	private DataElement addVariableIfNeeded(DataPetriNet net, DataAwareCausalNet dataCausalNet, String variable) {
		DataElement existingVariable = net.getVariable(variable);
		if (existingVariable == null) {
			existingVariable = net.addVariable(variable, dataCausalNet.getVariableType(variable), null, null);
		}
		return existingVariable;
	}

	private boolean isTauPlace(Place p) {
		Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> inEdges = p.getGraph().getInEdges(p);
		Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> outEdges = p.getGraph().getOutEdges(p);
		if (inEdges.size() == 1 && outEdges.size() == 1) {
			Transition source = (Transition) inEdges.iterator().next().getSource();
			Transition target = (Transition) outEdges.iterator().next().getTarget();
			return onlyOneOutgoing(source) && onlyOneIncoming(target) && source.isInvisible() && target.isInvisible();
		}
		return false;
	}

	private void removeFromNet(PetrinetGraph net, PetrinetNode node, Marking initialMarking, Marking finalMarking) {
		PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> source = FluentIterable.from(net.getInEdges(node))
				.first().get();
		PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> target = FluentIterable.from(net.getOutEdges(node))
				.first().get();
		if (node instanceof Transition) {
			net.removeTransition((Transition) node);
			mergePlaces(net, initialMarking, finalMarking, (Place) source.getSource(), (Place) target.getTarget());
		} else {
			net.removePlace((Place) node);
			mergeTransitions(net, (Transition) source.getSource(), (Transition) target.getTarget());
		}
	}

	private boolean isSimpleSequenceFlow(PetrinetNode t) {
		Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> inEdges = t.getGraph().getInEdges(t);
		Collection<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> outEdges = t.getGraph().getOutEdges(t);
		return inEdges.size() == 1 && outEdges.size() == 1 && onlyOneOutgoing(inEdges.iterator().next().getSource())
				&& onlyOneIncoming(outEdges.iterator().next().getTarget());
	}

	private boolean onlyOneIncoming(PetrinetNode node) {
		return node.getGraph().getInEdges(node).size() == 1;
	}

	private boolean onlyOneOutgoing(PetrinetNode node) {
		return node.getGraph().getOutEdges(node).size() == 1;
	}

	private void mergePlaces(PetrinetGraph net, Marking initialMarking, Marking finalMarking, Place source,
			Place target) {
		if (initialMarking.contains(source)) {
			for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge : net.getOutEdges(target)) {
				net.addArc(source, (Transition) edge.getTarget());
			}
			net.removePlace(target);

		} else {
			for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge : net.getInEdges(source)) {
				net.addArc((Transition) edge.getSource(), target);
			}
			net.removePlace(source);

		}
	}

	private void mergeTransitions(PetrinetGraph net, Transition source, Transition target) {
		for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge : net.getOutEdges(target)) {
			net.addArc(source, (Place) edge.getTarget());
		}
		net.removeTransition(target);
	}

	private boolean hasRelations(CNetNode currentTask, CausalNet causalNet) {
		return !causalNet.getInputBindings(currentTask).isEmpty()
				|| !causalNet.getOutputBindings(currentTask).isEmpty();
	}

	private Place addAnonymousPlace(PetrinetGraph net, LabelGenerator generator) {
		return net.addPlace(generator.newLabel());
	}

	private PNWDTransition addAnonymousTransition(DataPetriNet net, LabelGenerator generator) {
		PNWDTransition t = (PNWDTransition) net.addTransition(generator.newLabel());
		t.setInvisible(true);
		return t;
	}

	private Transition addTransitionIfNeeded(PetrinetGraph net, CNetNode node, Map<CNetNode, Transition> transitions) {
		Transition existingTransition = transitions.get(node);
		if (existingTransition == null) {
			Transition newTransition = net.addTransition(node.getLabel());
			transitions.put(node, newTransition);
			return newTransition;
		}
		return existingTransition;
	}

	private Place addPlaceIfNeeded(CNetNode task, Map<CNetNode, Place> places, PetrinetGraph net,
			LabelGenerator labelGenerator) {
		Place place = places.get(task);
		if (place == null) {
			place = addAnonymousPlace(net, labelGenerator);
			places.put(task, place);
		}
		return place;
	}

	private Place addGatewayIfNeeded(CNetNode source, CNetNode target, Map<Gateway, Place> gatewayPlaces,
			PetrinetGraph net, LabelGenerator labelGenerator) {
		Gateway gateway = new Gateway(source, target);
		Place place = gatewayPlaces.get(gateway);
		if (place == null) {
			place = addAnonymousPlace(net, labelGenerator);
			gatewayPlaces.put(gateway, place);
		}
		return place;
	}

	public boolean isReduceNet() {
		return isReduceNet;
	}

	public void setReduceNet(boolean isReduceNet) {
		this.isReduceNet = isReduceNet;
	}

	public LabelGenerator getTransitionLabelGenerator() {
		return transitionLabelGenerator;
	}

	public void setTransitionLabelGenerator(LabelGenerator transitionLabelGenerator) {
		this.transitionLabelGenerator = transitionLabelGenerator;
	}

	public LabelGenerator getPlaceLabelGenerator() {
		return placeLabelGenerator;
	}

	public void setPlaceLabelGenerator(LabelGenerator placeLabelGenerator) {
		this.placeLabelGenerator = placeLabelGenerator;
	}

	public BiMap<CNetBinding, Transition> getBindingToTransition() {
		return bindingToTransition;
	}

	public BiMap<CNetNode, Transition> getNodeToTransition() {
		return nodeToTransition;
	}

}
