package org.processmining.dataawarecnetminer.extension;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Set;

import org.processmining.framework.plugin.PluginContext;

import com.google.common.collect.Ordering;

/**
 * Generic registry for extensions
 * 
 * @author F. Mannhardt
 *
 */
public class HeuristicRegistry {

	private final static class DefaultOrder<T extends HeuristicExtension> implements Comparator<T> {
		public int compare(T o1, T o2) {
			DefaultHeuristicPlugin a1 = o1.getClass().getAnnotation(DefaultHeuristicPlugin.class);
			DefaultHeuristicPlugin a2 = o2.getClass().getAnnotation(DefaultHeuristicPlugin.class);
			if (a1 != null && a2 != null) {
				return 0;
			} else if (a1 != null) {
				// prefer default plugin
				return -1;
			} else if (a2 != null) {
				// prefer default plugin
				return 1;
			}
			return 0; // both null
		}
	}

	private final static class NameComparator<T extends HeuristicExtension> implements Comparator<T> {
		public int compare(T o1, T o2) {
			return o1.getName().compareTo(o2.getName());
		}
	}

	/**
	 * Returns all {@link HeuristicExtension} classes that are annotated with
	 * the {@link HeuristicPlugin} annotation and implement the class specified
	 * by parameter 'extensionClass'.
	 * 
	 * @param context
	 * @param extensionClass
	 * @return
	 */
	public <T extends HeuristicExtension> Iterable<T> getHeuristicExtensions(
			PluginContext context,
			Class<T> extensionClass) {
		Set<Class<?>> dependencyPlugins = context.getPluginManager()
				.getKnownClassesAnnotatedWith(HeuristicPlugin.class);
		List<T> extensions = new ArrayList<>();
		for (Class<?> pluginClass : dependencyPlugins) {
			tryAddExtension(extensions, extensionClass, pluginClass);
		}
		sortExtensions(extensions);
		return extensions;
	}

	@SuppressWarnings("unchecked")
	private <T extends HeuristicExtension> void tryAddExtension(List<T> extensions, Class<T> requestedClass,
			Class<?> foundClass) {
		try {
			if (requestedClass.isAssignableFrom(foundClass)) {
				extensions.add((T) foundClass.newInstance());
			}
		} catch (Exception e) {
			// ignore but log information, we cannot really do anything about this and it should not break the code if someone has a malformed plugin
			System.out.println("Could not load extension " + foundClass.getClass().getName() + ": " + e);
			e.printStackTrace();
		}
	}

	private <T extends HeuristicExtension> void sortExtensions(List<T> heuristicExtensions) {
		Ordering<T> extensionOrdering = Ordering.from(new DefaultOrder<T>()).compound(new NameComparator<T>());
		Collections.sort(heuristicExtensions, extensionOrdering);
	}

}