package org.processmining.dataawarecnetminer.extension.bindings;

import java.text.NumberFormat;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Set;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XTrace;
import org.processmining.dataawarecnetminer.common.MinerContext;
import org.processmining.dataawarecnetminer.exception.MiningException;
import org.processmining.dataawarecnetminer.model.EventRelationStorage;
import org.processmining.dataawarecnetminer.parallel.Command;
import org.processmining.dataawarecnetminer.parallel.TaskExecutor;
import org.processmining.models.causalgraph.SimpleCausalGraph;

import com.google.common.base.Joiner;
import com.google.common.collect.ConcurrentHashMultiset;
import com.google.common.collect.HashMultiset;
import com.google.common.collect.ImmutableMultiset;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Lists;
import com.google.common.collect.Multiset;
import com.google.common.collect.Multiset.Entry;
import com.google.common.collect.Multisets;

/**
 * Implements the nearest activity binding heuristic as proposed in the paper:
 * 
 * <pre>
 * AJMM Weijters, JTS Ribeiro. Flexible Heuristic Miner. DOI: 10.1109/CIDM.2011.5949453
 * </pre>
 * 
 * @author F. Mannhardt
 *
 */
public class BindingsNearestActivityImpl extends DefaultBindingsHeuristic {

	private Map<XEventClass, Multiset<Set<XEventClass>>> inputSets;
	private Map<XEventClass, Multiset<Set<XEventClass>>> outputSets;

	private SimpleCausalGraph causalGraph;
	private Collection<XTrace> traces;

	private Map<String, XEventClass> classById;
	private XEventClassifier classifier;

	private XEventClass startTask;
	private XEventClass endTask;

	public BindingsNearestActivityImpl(EventRelationStorage eventStorage) {
		super();
		this.classById = eventStorage.getEventClassesById();
		this.classifier = eventStorage.getClassifier();
		this.startTask = eventStorage.getStartEventClass();
		this.endTask = eventStorage.getEndEventClass();
		this.traces = eventStorage.getAllTraces();
	}

	@Override
	public void computeBindings(final MinerContext context, SimpleCausalGraph causalGraph) throws MiningException {
		this.causalGraph = causalGraph;
		this.inputSets = new HashMap<>();
		this.outputSets = new HashMap<>();
		for (XEventClass eventClass : classById.values()) {
			inputSets.put(eventClass, ConcurrentHashMultiset.<Set<XEventClass>>create());
			outputSets.put(eventClass, ConcurrentHashMultiset.<Set<XEventClass>>create());
		}
		try {
			TaskExecutor.Factory.getInstance().execute(context.getExecutor(), traces, new Command<XTrace>() {

				public void execute(XTrace trace) {
					addInputSet(inputSets, trace);
					addOutputSet(outputSets, trace);
				}
			});
		} catch (InterruptedException e) {
			Thread.currentThread().interrupt(); // stop and propagate
		}
	}

	private XEventClass getClassOf(XEvent event) {
		return classById.get(classifier.getClassIdentity(event));
	}

	private Map<XEventClass, Multiset<Set<XEventClass>>> addInputSet(
			Map<XEventClass, Multiset<Set<XEventClass>>> result, XTrace trace) {
		List<XEvent> reversedTrace = Lists.reverse(trace);
		ListIterator<XEvent> iterator = reversedTrace.listIterator();

		registerInputSetActivations(result, iterator, iterator.nextIndex(), endTask);
		while (iterator.hasNext()) {
			XEvent event = iterator.next();
			registerInputSetActivations(result, iterator, iterator.nextIndex(), getClassOf(event));
		}
		getOrCreateSet(result, startTask); // always empty
		return result;
	}

	private Map<XEventClass, Multiset<Set<XEventClass>>> addOutputSet(
			Map<XEventClass, Multiset<Set<XEventClass>>> result, XTrace trace) {
		ListIterator<XEvent> iterator = trace.listIterator();

		registerOutputSetActivations(result, iterator, iterator.nextIndex(), startTask);
		while (iterator.hasNext()) {
			XEvent event = iterator.next();
			registerOutputSetActivations(result, iterator, iterator.nextIndex(), getClassOf(event));
		}
		getOrCreateSet(result, endTask); // always empty
		return result;
	}

	private void registerInputSetActivations(Map<XEventClass, Multiset<Set<XEventClass>>> inputSets,
			ListIterator<XEvent> iterator, int nextIndex, XEventClass currentTask) {
		Set<XEventClass> synchronizedTasks = getSynchronizedInputTasks(iterator, nextIndex, currentTask);
		Multiset<Set<XEventClass>> inputSet = getOrCreateSet(inputSets, currentTask);
		if (!synchronizedTasks.isEmpty()) {
			inputSet.add(synchronizedTasks);
		}
	}

	private void registerOutputSetActivations(Map<XEventClass, Multiset<Set<XEventClass>>> result,
			ListIterator<XEvent> iterator, int nextIndex, XEventClass currentTask) {
		Set<XEventClass> activatedTasks = getActivatedOutputTasks(iterator, nextIndex, currentTask);
		Multiset<Set<XEventClass>> outputSet = getOrCreateSet(result, currentTask);
		if (!activatedTasks.isEmpty()) {
			outputSet.add(activatedTasks);
		}
	}

	private Multiset<Set<XEventClass>> getOrCreateSet(Map<XEventClass, Multiset<Set<XEventClass>>> taskToSet,
			XEventClass currentTask) {
		Multiset<Set<XEventClass>> outputSet = taskToSet.get(currentTask);
		if (outputSet == null) {
			outputSet = HashMultiset.create();
			taskToSet.put(currentTask, outputSet);
		}
		return outputSet;
	}

	private Set<XEventClass> getSynchronizedInputTasks(ListIterator<XEvent> iterator, int nextIndex,
			XEventClass currentTask) {
		Set<XEventClass> synchronizedTasks = new HashSet<>();
		Set<XEventClass> inputTasks = causalGraph.getPreset(currentTask);

		// for each inputTask in the dependency graph
		for (XEventClass inputTask : inputTasks) {

			// look whether there is no other possibly synchronizing task (outputTask of the inputTask) between currentTask and inputTask
			Set<XEventClass> outputTasks = causalGraph.getPostset(inputTask);
			if (isNearestCandidateTask(iterator, inputTask, outputTasks)) {
				// this outputTask was activated by currentTask as there is no other task from the inputSet in between and the outputTask is present
				synchronizedTasks.add(inputTask);
			}
			for (; iterator.nextIndex() != nextIndex; iterator.previous()) {
			}

		}
		return synchronizedTasks;
	}

	private Set<XEventClass> getActivatedOutputTasks(ListIterator<XEvent> iterator, int nextIndex,
			XEventClass currentTask) {
		Set<XEventClass> activatedTasks = new HashSet<>();
		Set<XEventClass> outputTasks = causalGraph.getPostset(currentTask);

		// for each outputTask in the dependency graph
		for (XEventClass outputTask : outputTasks) {

			// look whether there is no other possibly activating task between task and outputTask
			Set<XEventClass> inputTasks = causalGraph.getPreset(outputTask);
			if (isNearestCandidateTask(iterator, outputTask, inputTasks)) {
				// this outputTask was activated by currentTask as there is no other task from the inputSet in between and the outputTask is present
				activatedTasks.add(outputTask);
			}
			// reset iterator for next task
			for (; iterator.nextIndex() != nextIndex; iterator.previous()) {
			}
		}
		return activatedTasks;
	}

	private boolean isNearestCandidateTask(ListIterator<XEvent> iterator, XEventClass taskToFind,
			Set<XEventClass> candidateSet) {

		while (iterator.hasNext()) {
			XEvent candidateEvent = iterator.next();
			XEventClass candidate = getClassOf(candidateEvent);
			if (candidate.equals(taskToFind)) {
				return true; // we found the task without any candidate in between
			} else if (candidateSet.contains(candidate)) {
				// found another candidate beforehand
				return false;
			}
		}

		return taskToFind.equals(endTask) || taskToFind.equals(startTask); // the trace does not contain the taskToFind or we activated the endTask or the startTask
	}

	@Override
	public Bindings getFilteredBindings(final BindingsHeuristicConfig bindingsConfig) {

		final double bindingsThreshold = bindingsConfig.getBindingsThreshold();

		final Map<XEventClass, Multiset<Set<XEventClass>>> filteredInputSets = filterInputSets(bindingsThreshold);
		final Map<XEventClass, Multiset<Set<XEventClass>>> filteredOutputSets = filterOutputSets(bindingsThreshold);

		return new Bindings() {

			public Multiset<Set<XEventClass>> getOutputSets(XEventClass task) {
				return filteredOutputSets.get(task);
			}

			public Multiset<Set<XEventClass>> getInputSets(XEventClass task) {
				return filteredInputSets.get(task);
			}

			public String toString() {
				StringBuilder stringBuilder = new StringBuilder();
				stringBuilder.append(
						"Binding Threshold: " + NumberFormat.getNumberInstance().format(bindingsThreshold) + "\n");
				stringBuilder.append("Input Bindings:\n");
				Joiner.on('\n').withKeyValueSeparator("=").appendTo(stringBuilder, filteredInputSets);
				stringBuilder.append("\n\nOutput Bindings:\n");
				Joiner.on('\n').withKeyValueSeparator("=").appendTo(stringBuilder, filteredOutputSets);
				return stringBuilder.toString();
			}

		};
	}

	private Map<XEventClass, Multiset<Set<XEventClass>>> filterInputSets(final double bindingsThreshold) {
		final Map<XEventClass, Multiset<Set<XEventClass>>> filteredInputSets = new HashMap<>();
		for (java.util.Map.Entry<XEventClass, Multiset<Set<XEventClass>>> entry : inputSets.entrySet()) {
			filteredInputSets.put(entry.getKey(),
					filterAndEnsureConnected(bindingsThreshold, entry, causalGraph.getPreset(entry.getKey())));
		}
		return filteredInputSets;
	}

	private Map<XEventClass, Multiset<Set<XEventClass>>> filterOutputSets(final double bindingsThreshold) {
		final Map<XEventClass, Multiset<Set<XEventClass>>> filteredOutputSets = new HashMap<>();
		for (java.util.Map.Entry<XEventClass, Multiset<Set<XEventClass>>> entry : outputSets.entrySet()) {
			filteredOutputSets.put(entry.getKey(),
					filterAndEnsureConnected(bindingsThreshold, entry, causalGraph.getPostset(entry.getKey())));
		}
		return filteredOutputSets;
	}

	private boolean anySubsetContains(Multiset<Set<XEventClass>> multiset, XEventClass obj) {
		for (Set<XEventClass> element : multiset.elementSet()) {
			if (element.contains(obj)) {
				return true;
			}
		}
		return false;
	}

	private Multiset<Set<XEventClass>> filterAndEnsureConnected(final double bindingsThreshold,
			java.util.Map.Entry<XEventClass, Multiset<Set<XEventClass>>> bindings,
			Set<XEventClass> requiredActivities) {
	
		ImmutableMultiset<Set<XEventClass>> sortedBindings = Multisets.copyHighestCountFirst(bindings.getValue());
		Multiset<Set<XEventClass>> filteredBindings = filterBindings(sortedBindings, bindingsThreshold);

		for (XEventClass activity : requiredActivities) {
			ensureConnected(sortedBindings, filteredBindings, activity);
		}

		return filteredBindings;
	}

	private void ensureConnected(ImmutableMultiset<Set<XEventClass>> sortedBindings,
			Multiset<Set<XEventClass>> filteredBindings, XEventClass activity) {
		if (!anySubsetContains(filteredBindings, activity)) {
			for (Entry<Set<XEventClass>> binding : sortedBindings.entrySet()) {
				if (binding.getElement().contains(activity)) {
					// best binding containing the activity
					filteredBindings.add(binding.getElement(), binding.getCount());
					return; // we are done with this one
				}
			}
			// no binding found that the activity was part of, 
			// add it anyway because a relation without binding does not make sense
			filteredBindings.add(ImmutableSet.of(activity));
		}
	}

	/**
	 * Filters a multiset of bindings according to their frequency of
	 * occurrence. Then, checks whether all connected activities are present in
	 * a binding. If not, then, it adds the best binding such that all conencted
	 * activities are at least in one binding.
	 * 
	 * @param bindings
	 * @param bindingThreshold
	 * @return
	 */
	private Multiset<Set<XEventClass>> filterBindings(Multiset<Set<XEventClass>> sortedBindings,
			double bindingThreshold) {
		Multiset<Set<XEventClass>> filteredMultiset = HashMultiset.create();
		int maxBindingCount = -1;
		for (Entry<Set<XEventClass>> entry : sortedBindings.entrySet()) {
			Set<XEventClass> binding = entry.getElement();
			int bindingCount = entry.getCount();
			maxBindingCount = Math.max(maxBindingCount, bindingCount);
			double countRatio = bindingCount / (double) maxBindingCount;
			if (countRatio >= bindingThreshold && !filteredMultiset.contains(binding)) {
				filteredMultiset.add(binding, bindingCount);
			}
		}
		return filteredMultiset;
	}

}