package org.processmining.dataawarecnetminer.extension.conformance;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ExecutorService;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.dataawarecnetminer.common.MinerContext;
import org.processmining.dataawarecnetminer.converter.CausalNetToPetrinet;
import org.processmining.dataawarecnetminer.exception.MiningException;
import org.processmining.dataawarecnetminer.model.EventRelationStorage;
import org.processmining.dataawarecnetminer.model.FrequencyAwareCausalNet;
import org.processmining.dataawarecnetminer.model.FrequencyAwareDataCausalNet;
import org.processmining.dataawarecnetminer.model.FrequencyAwareNet;
import org.processmining.dataawarecnetminer.parallel.Command;
import org.processmining.dataawarecnetminer.parallel.TaskExecutor;
import org.processmining.datapetrinets.DataPetriNetsWithMarkings;
import org.processmining.models.cnet.CNetBinding;
import org.processmining.models.cnet.CNetNode;
import org.processmining.models.cnet.CausalNet;
import org.processmining.models.dcnet.DCNet;
import org.processmining.models.dcnet.DataAwareCausalNet;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.plugins.balancedconformance.BalancedDataXAlignmentPlugin;
import org.processmining.plugins.balancedconformance.config.BalancedProcessorConfiguration;
import org.processmining.plugins.balancedconformance.controlflow.ControlFlowAlignmentException;
import org.processmining.plugins.balancedconformance.dataflow.exception.DataAlignmentException;
import org.processmining.xesalignmentextension.XAlignmentExtension;
import org.processmining.xesalignmentextension.XAlignmentExtension.MoveType;
import org.processmining.xesalignmentextension.XAlignmentExtension.XAlignedLog;
import org.processmining.xesalignmentextension.XAlignmentExtension.XAlignment;
import org.processmining.xesalignmentextension.XAlignmentExtension.XAlignmentMove;
import org.processmining.xesalignmentextension.XDataAlignmentExtension;
import org.processmining.xesalignmentextension.XDataAlignmentExtension.DataMoveType;
import org.processmining.xesalignmentextension.XDataAlignmentExtension.XDataAlignmentMove;

import com.google.common.collect.BiMap;
import com.google.common.collect.ConcurrentHashMultiset;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Multiset;

public final class ConformanceAlignmentImpl extends DefaultConformanceHeuristic {

	public static class FrequencyAwareDataNetAlignmentImpl implements FrequencyAwareNet {

		private final Multiset<CNetNode> nodeFrequency = ConcurrentHashMultiset.create();
		private final Multiset<CNetNode> nodeSkippedFrequency = ConcurrentHashMultiset.create();
		private final Multiset<CNetNode> nodeInsertedFrequency = ConcurrentHashMultiset.create();
		private final Map<CNetNode, Multiset<String>> variableCorrectFrequency = new HashMap<>();
		private final Map<CNetNode, Multiset<String>> variableWrongFrequency = new HashMap<>();
		private final Map<CNetNode, Multiset<String>> variableMissingFrequency = new HashMap<>();
		private final Multiset<CNetBinding> bindingFrequency = ConcurrentHashMultiset.create();

		public FrequencyAwareDataNetAlignmentImpl(XAlignedLog alignedLog, final DataAwareCausalNet dcNet,
				final BiMap<CNetBinding, Transition> bindingMap, final BiMap<CNetNode, Transition> nodeMap,
				final Map<String, Transition> transitionsLocalId, ExecutorService executorService) {

			final CausalNet cNet = dcNet.getCausalNet();

			if (cNet.getStartNode() != null) {
				nodeFrequency.add(cNet.getStartNode(), alignedLog.size());
			}
			if (cNet.getEndNode() != null) {
				nodeFrequency.add(cNet.getEndNode(), alignedLog.size());
			}

			for (CNetNode node : cNet.getNodes()) {
				variableCorrectFrequency.put(node, ConcurrentHashMultiset.<String>create());
				variableWrongFrequency.put(node, ConcurrentHashMultiset.<String>create());
				variableMissingFrequency.put(node, ConcurrentHashMultiset.<String>create());
			}

			final Map<String, CNetNode> nameToNode = createNodeNameLookup(cNet);
			try {
				TaskExecutor.Factory.getInstance().execute(executorService, alignedLog, new Command<XAlignment>() {

					public void execute(XAlignment alignment) {
						for (XAlignmentMove move : alignment) {
							Transition transition = transitionsLocalId.get(move.getActivityId());
							CNetBinding binding = bindingMap.inverse().get(transition);
							if (binding != null) {
								bindingFrequency.add(binding);
							} else {
								recordActivityDeviations(cNet, nodeMap, nameToNode, move, transition);
							}
						}
					}

					private void recordActivityDeviations(final CausalNet cNet,
							final BiMap<CNetNode, Transition> nodeMap, final Map<String, CNetNode> nameToNode,
							XAlignmentMove move, Transition transition) {
						MoveType type = move.getType();
						if (type == MoveType.MODEL || type == MoveType.SYNCHRONOUS) {
							CNetNode node = nodeMap.inverse().get(transition);
							if (node != null) {
								nodeFrequency.add(node);
								if (type == MoveType.MODEL) {
									nodeInsertedFrequency.add(node);
								}
								recordDataDeviations(cNet, move, node);
							}
						} else { // log move
							CNetNode matchedNode = nameToNode.get(move.getLogMove());
							if (matchedNode != null) {
								nodeFrequency.add(matchedNode);
								nodeSkippedFrequency.add(matchedNode);
							}
						}
					}

					private void recordDataDeviations(final CausalNet cNet, XAlignmentMove move, CNetNode node) {
						Set<String> variables = dcNet.getWrites(node);
						if (!variables.isEmpty()) {
							XDataAlignmentMove dataMove = XDataAlignmentExtension.instance().extendXAlignmentMove(move);
							if (dataMove.getDataMoveType() == DataMoveType.CORRECT) {
								variableCorrectFrequency.get(node).addAll(variables);
							} else {
								// some are incorrect / missing
								XEvent event = dataMove.getEvent();
								for (String variable : variables) {
									XAttribute attribute = event.getAttributes().get(variable);
									if (attribute != null) {
										DataMoveType deviationType = XDataAlignmentExtension.instance()
												.extractDataMoveType(attribute);
										if (deviationType != null) {
											switch (deviationType) {
												case INCORRECT :
													variableWrongFrequency.get(node).add(variable);
													break;
												case MISSING :
													variableMissingFrequency.get(node).add(variable);
													break;
												case CORRECT :
												default :
													variableCorrectFrequency.get(node).add(variable);
													break;
											}
										}
									} else {
										// should not happen but would be same as missing
										variableMissingFrequency.get(node).addAll(variables);
									}
								}
							}
						}
					}
				});
			} catch (InterruptedException e) {
				Thread.currentThread().interrupt();
			}
		}

		private Map<String, CNetNode> createNodeNameLookup(CausalNet cNet) {
			final Map<String, CNetNode> nameToNode = new HashMap<>();
			for (CNetNode node : cNet.getNodes()) {
				nameToNode.put(node.getLabel(), node);
			}
			return nameToNode;
		}

		public int getOutputFrequency(CNetNode node) {
			return nodeFrequency.count(node);
		}

		public int getNodeFrequency(CNetNode node) {
			return nodeFrequency.count(node);
		}

		public int getInputFrequency(CNetNode node) {
			return nodeFrequency.count(node);
		}

		public int getBindingFrequency(CNetBinding binding) {
			return bindingFrequency.count(binding);
		}

		public int getNodeSkippedFrequency(CNetNode node) {
			return nodeSkippedFrequency.count(node);
		}

		public int getNodeInsertedFrequency(CNetNode node) {
			return nodeInsertedFrequency.count(node);
		}

		public int getIncorrectVariableFrequency(CNetNode node, String variable) {
			return nullSafeCount(variable, variableWrongFrequency.get(node));
		}

		public int getCorrectVariableFrequency(CNetNode node, String variable) {
			return nullSafeCount(variable, variableCorrectFrequency.get(node));
		}

		public int getMissingVariableFrequency(CNetNode node, String variable) {
			return nullSafeCount(variable, variableMissingFrequency.get(node));
		}

		private int nullSafeCount(String variable, Multiset<String> multiset) {
			if (multiset != null) {
				return multiset.count(variable);
			} else {
				return 0;
			}
		}

	}

	public final static class FrequencyAwareDataCausalNetAlignmentImpl extends FrequencyAwareDataNetAlignmentImpl
			implements FrequencyAwareDataCausalNet {

		private final DataAwareCausalNet dcNet;

		private FrequencyAwareDataCausalNetAlignmentImpl(XAlignedLog alignedLog, DataAwareCausalNet dcNet,
				BiMap<CNetBinding, Transition> bindingMap, BiMap<CNetNode, Transition> nodeMap,
				Map<String, Transition> transitionsLocalId, ExecutorService executorService) {
			super(alignedLog, dcNet, bindingMap, nodeMap, transitionsLocalId, executorService);
			this.dcNet = dcNet;
		}

		public DataAwareCausalNet getDCNet() {
			return dcNet;
		}

	}

	private final static class FrequencyAwareCausalNetAlignmentImpl extends FrequencyAwareDataNetAlignmentImpl
			implements FrequencyAwareCausalNet {

		private final CausalNet cNet;

		private FrequencyAwareCausalNetAlignmentImpl(XAlignedLog alignedLog, CausalNet cNet,
				BiMap<CNetBinding, Transition> bindingMap, BiMap<CNetNode, Transition> nodeMap,
				Map<String, Transition> transitionsLocalId, ExecutorService executorService) {
			super(alignedLog, new DCNet(cNet), bindingMap, nodeMap, transitionsLocalId, executorService);
			this.cNet = cNet;
		}

		public CausalNet getCNet() {
			return cNet;
		}
	}

	private final EventRelationStorage eventStorage;

	public ConformanceAlignmentImpl(EventRelationStorage eventStorage) {
		super();
		this.eventStorage = eventStorage;
	}

	public FrequencyAwareCausalNet computeFrequencies(MinerContext context, FrequencyAwareCausalNet cnet)
			throws MiningException {
		CausalNetToPetrinet converter = new CausalNetToPetrinet();
		converter.setReduceNet(false);
		DataPetriNetsWithMarkings pn = converter.convert(DataAwareCausalNet.Factory.fromCausalNet(cnet.getCNet()));

		Map<String, Transition> transitionsLocalId = createTransitionMap(pn);
		BalancedProcessorConfiguration config = createConfig(pn, ImmutableSet.<String>of());
		XAlignedLog alignedLog = alignTraces(context, pn, config);

		return new FrequencyAwareCausalNetAlignmentImpl(alignedLog, cnet.getCNet(), converter.getBindingToTransition(),
				converter.getNodeToTransition(), transitionsLocalId, context.getExecutor());
	}

	public FrequencyAwareDataCausalNet computeFrequencies(MinerContext context, FrequencyAwareDataCausalNet dcNet)
			throws MiningException {
		CausalNetToPetrinet converter = new CausalNetToPetrinet();
		converter.setReduceNet(false);
		DataPetriNetsWithMarkings pn = converter.convert(dcNet.getDCNet());

		BalancedProcessorConfiguration config = createConfig(pn, dcNet.getDCNet().getVariables());

		Map<String, Transition> transitionsLocalId = createTransitionMap(pn);
		XAlignedLog alignedLog = alignTraces(context, pn, config);

		return new FrequencyAwareDataCausalNetAlignmentImpl(alignedLog, dcNet.getDCNet(),
				converter.getBindingToTransition(), converter.getNodeToTransition(), transitionsLocalId,
				context.getExecutor());
	}

	private BalancedProcessorConfiguration createConfig(DataPetriNetsWithMarkings pn, Set<String> variables) {
		XEventClasses eventClasses = createEventClasses();
		BalancedProcessorConfiguration config = BalancedProcessorConfiguration.newDefaultInstance(pn,
				pn.getInitialMarking(), pn.getFinalMarkings(), eventClasses, variables, 2, 2, 1, 1);
		BalancedProcessorConfiguration.autoGuessBounds(config, pn, getAsLog(eventStorage.getAllTraces()));
		return config;
	}
	
	private static XLog getAsLog(Collection<XTrace> traces) {
		if (traces instanceof XLog) {
			return (XLog) traces;
		} else {
			XLog log = XFactoryRegistry.instance().currentDefault().createLog();
			log.addAll(traces);
			return log;
		}
	}

	private XAlignedLog alignTraces(MinerContext context, DataPetriNetsWithMarkings pn,
			BalancedProcessorConfiguration config) throws MiningException {
		context.getProgress().setIndeterminate(false);

		XAlignedLog alignedLog;
		try {
			BalancedDataXAlignmentPlugin alignmentPlugin = new BalancedDataXAlignmentPlugin();
			alignedLog = XAlignmentExtension.instance().extendLog(
					alignmentPlugin.alignLog(context.getProgress(), pn, eventStorage.getAllTraces(), config));

		} catch (ControlFlowAlignmentException | DataAlignmentException e) {
			throw new MiningException(e);
		}
		return alignedLog;
	}

	private XEventClasses createEventClasses() {
		XEventClasses eventClasses = new XEventClasses(eventStorage.getClassifier());
		for (XEventClass eventClass : eventStorage.getEventClasses()) {
			if (!(eventClass.equals(eventStorage.getStartEventClass())
					|| eventClass.equals(eventStorage.getEndEventClass()))) {
				eventClasses.register(eventClass.getId());
			}
		}
		return eventClasses;
	}

	private Map<String, Transition> createTransitionMap(DataPetriNetsWithMarkings pn) {
		Map<String, Transition> transitionsLocalId = new HashMap<>();
		for (Transition t : pn.getTransitions()) {
			transitionsLocalId.put(t.getLocalID().toString(), t);
		}
		return transitionsLocalId;
	}

}