package org.processmining.dataawarecnetminer.interactive.util;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.GraphicsDevice;
import java.awt.GraphicsEnvironment;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.io.IOException;
import java.lang.reflect.Field;
import java.text.DateFormat;
import java.util.Date;
import java.util.Map;
import java.util.Set;

import javax.swing.AbstractAction;
import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.KeyStroke;

import org.deckfour.xes.classification.XEventClass;
import org.processmining.dataawarecnetminer.model.EventRelationStorage;
import org.processmining.framework.util.ui.widgets.helper.ProMUIHelper;
import org.processmining.models.causalgraph.Relation;
import org.processmining.plugins.graphviz.dot.Dot;
import org.processmining.plugins.graphviz.dot.DotEdge;
import org.processmining.plugins.graphviz.dot.DotElement;
import org.processmining.plugins.graphviz.dot.DotNode;
import org.processmining.plugins.graphviz.visualisation.DotPanel;
import org.processmining.plugins.graphviz.visualisation.NavigableSVGPanel;
import org.processmining.plugins.graphviz.visualisation.listeners.SelectionChangedListener;

import com.kitfox.svg.Group;
import com.kitfox.svg.SVGDiagram;
import com.kitfox.svg.SVGElement;
import com.kitfox.svg.SVGUniverse;

public class DotHelper {

	private static final class FreezeGraphAction extends AbstractAction {

		private static final long serialVersionUID = -1540697035814396137L;

		private final DotPanel dotPanel;

		FreezeGraphAction(DotPanel dotPanel) {
			super();
			this.dotPanel = dotPanel;
		}

		public void actionPerformed(ActionEvent e) {
			NavigableSVGPanel svgPanel;
			try {

				SVGDiagram svgDiagram = dotPanel.getSVG();
				SVGUniverse universe = svgDiagram.getUniverse();
				Field field = universe.getClass().getDeclaredField("cachedReader");
				field.setAccessible(true);
				field.set(universe, null);

				SVGUniverse frozenUniverse = universe.duplicate();
				svgPanel = new NavigableSVGPanel(frozenUniverse.getDiagram(dotPanel.getSVG().getXMLBase()));
				JFrame frame = new JFrame();
				frame.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
				frame.setTitle("Snapshot: " + DateFormat.getDateTimeInstance().format(new Date()));
				frame.setBackground(Color.WHITE);
				frame.add(svgPanel);
				frame.pack();
				showFrame(frame);
			} catch (ClassNotFoundException | IOException | IllegalArgumentException | IllegalAccessException
					| NoSuchFieldException | SecurityException e1) {
				ProMUIHelper.showErrorMessage(dotPanel, e1.toString(), "Error duplicating SVG");
			}
		}

		private void showFrame(JFrame frame) {
			GraphicsEnvironment ge = GraphicsEnvironment.getLocalGraphicsEnvironment();
			GraphicsDevice[] gs = ge.getScreenDevices();
			if (gs.length > 1) {
				for (int i = 0; i < gs.length; i++) {
					if (gs[i] != dotPanel.getGraphicsConfiguration().getDevice()) {
						JFrame dummy = new JFrame(gs[i].getDefaultConfiguration());
						frame.setLocationRelativeTo(dummy);
						frame.setExtendedState(Frame.MAXIMIZED_BOTH);
						frame.setAlwaysOnTop(true);
						dummy.dispose();
						break;
					}
				}
			} else {
				Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
				frame.setLocation(screenSize.width / 6, screenSize.height / 6);
				frame.setSize(screenSize.width * 2 / 3, screenSize.height * 2 / 3);
				frame.setExtendedState(Frame.NORMAL);
			}
			frame.setVisible(true);
		}
	}

	private int maxEdgeWidth = 5;
	
	public DotHelper() {
	}

	public DotPanel createPanel(Dot dot) {

		final DotPanel dotPanel = new DotPanel(dot);
		FreezeGraphAction action = new FreezeGraphAction(dotPanel);

		dotPanel.getHelperControlsExplanations().add("freeze in new window");
		dotPanel.getHelperControlsShortcuts().add("ctrl n");
		dotPanel.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW)
				.put(KeyStroke.getKeyStroke(KeyEvent.VK_N, InputEvent.CTRL_MASK), "newWindow");
		dotPanel.getActionMap().put("newWindow", action);

		dotPanel.addSelectionChangedListener(new SelectionChangedListener<DotElement>() {

			public void selectionChanged(Set<DotElement> selectedElements) {
				if (selectedElements.isEmpty()) {
					for (DotEdge edge : dotPanel.getDot().getEdges()) {
						styleEdge(edge, dotPanel.getSVG(), Color.BLACK, false);
					}
					for (DotNode node : dotPanel.getDot().getNodes()) {
						styleNode(node, dotPanel.getSVG(), Color.BLACK, false);
					}
				} else {
					for (DotEdge edge : dotPanel.getDot().getEdges()) {
						styleEdge(edge, dotPanel.getSVG(), Color.BLACK, true);
					}
					for (DotNode node : dotPanel.getDot().getNodes()) {
						styleNode(node, dotPanel.getSVG(), Color.BLACK, true);
					}
					for (DotNode node : dotPanel.getDot().getNodes()) {
						if (selectedElements.contains(node)) {
							styleNode(node, dotPanel.getSVG(), Color.BLACK, false);
							for (DotEdge edge : dotPanel.getDot().getEdges()) {
								if (edge.getSource().equals(node)) {
									styleEdge(edge, dotPanel.getSVG(), Color.BLACK, false);
									styleNode(edge.getTarget(), dotPanel.getSVG(), Color.BLACK, false);
								}
								if (edge.getTarget().equals(node)) {
									styleEdge(edge, dotPanel.getSVG(), Color.BLACK, false);
									styleNode(edge.getSource(), dotPanel.getSVG(), Color.BLACK, false);
								}
							}
						}
					}
				}
				dotPanel.repaint();
			}

			private void styleNode(DotNode node, SVGDiagram svg, Color color, boolean hide) {

				//prepare parts of the rendered dot element
				Group group = DotPanel.getSVGElementOf(svg, node);

				//transparency
				if (hide) {
					DotPanel.setCSSAttributeOf(group, "opacity", "0.3");
				} else {
					DotPanel.setCSSAttributeOf(group, "opacity", "1.0");
				}

			}

			private void styleEdge(DotEdge edge, SVGDiagram svg, Color color, boolean hide) {

				//prepare parts of the rendered dot element
				Group group = DotPanel.getSVGElementOf(svg, edge);
				SVGElement line = group.getChild(1);
				SVGElement arrowHead = group.getChild(2);

				//stroke
				DotPanel.setCSSAttributeOf(line, "stroke", color);
				DotPanel.setCSSAttributeOf(arrowHead, "stroke", color);
				DotPanel.setCSSAttributeOf(arrowHead, "fill", color);

				//transparency
				if (hide) {
					DotPanel.setCSSAttributeOf(group, "opacity", "0.1");
				} else {
					DotPanel.setCSSAttributeOf(group, "opacity", "1.0");
				}

			}

		});

		return dotPanel;
	}

	public DotNode addNodeIfNeeded(EventRelationStorage dfg, Dot dot, XEventClass eventClass,
			Map<XEventClass, DotNode> activityToNode) {
		DotNode existingNode = activityToNode.get(eventClass);
		if (existingNode == null) {
			if (eventClass.equals(dfg.getStartEventClass())) {
				DotNode startNode = dot.addNode("");
				startNode.setOption("shape", "box");
				startNode.setOption("style", "filled");
				startNode.setOption("height", ".1");
				startNode.setOption("width", ".4");
				startNode.setOption("fillcolor", "black");
				startNode.setSelectable(true);
				activityToNode.put(eventClass, startNode);
				return startNode;
			} else if (eventClass.equals(dfg.getEndEventClass())) {
				DotNode endNode = dot.addNode("");
				endNode.setOption("shape", "box");
				endNode.setOption("style", "filled");
				endNode.setOption("height", ".1");
				endNode.setOption("width", ".4");
				endNode.setOption("fillcolor", "black");
				endNode.setSelectable(true);
				activityToNode.put(eventClass, endNode);
				return endNode;
			} else {
				DotNode newNode = dot.addNode(eventClass.getId());
				newNode.setOption("shape", "box");
				newNode.setSelectable(true);
				activityToNode.put(eventClass, newNode);
				return newNode;
			}
		} else {
			return existingNode;
		}
	}

	public DotEdge addDependency(EventRelationStorage dfg, Dot dot, Map<XEventClass, DotNode> activityToNode,
			Relation relation) {
		DotNode sourceNode = addNodeIfNeeded(dfg, dot, relation.getSource(), activityToNode);
		DotNode targetNode = addNodeIfNeeded(dfg, dot, relation.getTarget(), activityToNode);
		return dot.addEdge(sourceNode, targetNode);
	}

	public double calcEdgeWidth(double value) {
		return 0.5 + value * maxEdgeWidth;
	}

	public double calcEdgeWidth(double count, double maxCount) {
		return 0.5 + (count / maxCount) * maxEdgeWidth;
	}

}
