package org.processmining.dataawarecnetminer.mining.classic;

import java.text.MessageFormat;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.model.XLog;
import org.processmining.dataawarecnetminer.common.MinerContext;
import org.processmining.dataawarecnetminer.common.ProgressNoOpImpl;
import org.processmining.dataawarecnetminer.converter.CausalToCNet;
import org.processmining.dataawarecnetminer.exception.MiningException;
import org.processmining.dataawarecnetminer.extension.bindings.Bindings;
import org.processmining.dataawarecnetminer.extension.bindings.BindingsHeuristic;
import org.processmining.dataawarecnetminer.extension.bindings.BindingsHeuristicConfig;
import org.processmining.dataawarecnetminer.extension.bindings.BindingsNearestActivityImpl;
import org.processmining.dataawarecnetminer.model.DependencyRelation;
import org.processmining.dataawarecnetminer.model.EventRelationStorage;
import org.processmining.dataawarecnetminer.model.FrequencyAwareCausalNet;
import org.processmining.dataawarecnetminer.model.FrequencyAwareCausalNetImpl;
import org.processmining.dataawarecnetminer.util.GraphUtils;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.Progress;
import org.processmining.models.causalgraph.Relation;
import org.processmining.models.causalgraph.SimpleCausalGraph;
import org.processmining.models.causalgraph.SimpleCausalGraphImpl;
import org.processmining.models.cnet.CausalNet;

import com.google.common.collect.ImmutableSet;
import com.google.common.collect.ImmutableSet.Builder;
import com.google.common.collect.Sets;

public class HeuristicsCausalNetMiner {

	public static class CausalNetConfig {

		private double bindingsThreshold = 0.1;
		private double longTermDependencyThreshold = 0.8;
		private boolean noTasksUnconnected = false;
		private boolean considerLongDistanceRelations = false;

		public boolean isConsiderLongDistanceRelations() {
			return considerLongDistanceRelations;
		}

		public void setConsiderLongDistanceRelations(boolean considerLongDistanceRelations) {
			this.considerLongDistanceRelations = considerLongDistanceRelations;
		}

		public double getBindingsThreshold() {
			return bindingsThreshold;
		}

		public void setBindingsThreshold(double bindingsThreshold) {
			this.bindingsThreshold = bindingsThreshold;
		}

		public double getLongTermDependencyThreshold() {
			return longTermDependencyThreshold;
		}

		public void setLongTermDependencyThreshold(double longTermDependencyThreshold) {
			this.longTermDependencyThreshold = longTermDependencyThreshold;
		}

		public boolean isNoTasksUnconnected() {
			return noTasksUnconnected;
		}

		public void setNoTasksUnconnected(boolean noTasksUnconnected) {
			this.noTasksUnconnected = noTasksUnconnected;
		}

	}

	private final EventRelationStorage eventRelations;
	private CausalNetConfig config = new CausalNetConfig();

	public HeuristicsCausalNetMiner(EventRelationStorage eventRelations) {
		this.eventRelations = eventRelations;
	}

	public HeuristicsCausalNetMiner(XLog log, XEventClassifier classifier) {
		this(EventRelationStorage.Factory.createEventRelations(log, classifier));
	}

	public FrequencyAwareCausalNet mineCausalNet(SimpleCausalGraph causalGraph) throws MiningException {
		ExecutorService executor = Executors.newFixedThreadPool(Runtime.getRuntime().availableProcessors());
		try {
			return mineCausalNet(causalGraph, executor);
		} finally {
			executor.shutdown();
		}
	}

	public FrequencyAwareCausalNet mineCausalNet(SimpleCausalGraph causalGraph, final ExecutorService executor)
			throws MiningException {
		return mineCausalNet(new MinerContext() {

			public Progress getProgress() {
				return new ProgressNoOpImpl();
			}

			public ExecutorService getExecutor() {
				return executor;
			}

			public PluginContext getPluginContext() {
				return null;
			}
		}, causalGraph, new BindingsNearestActivityImpl(eventRelations));
	}

	public FrequencyAwareCausalNet mineCausalNet(MinerContext context, SimpleCausalGraph causalGraph,
			BindingsHeuristic bindingsHeuristic) throws MiningException {

		checkHasSourcesAndSinks(causalGraph);

		if (!causalGraph.getSetActivities().contains(eventRelations.getStartEventClass())
				|| !causalGraph.getSetActivities().contains(eventRelations.getEndEventClass())) {
			causalGraph = addStartEndActivities(causalGraph, eventRelations.getStartEventClass(),
					eventRelations.getEndEventClass());
		}

		checkSameEventClasses(causalGraph);

		if (config.isConsiderLongDistanceRelations()) {

			Bindings bindings = calculateBindings(context, causalGraph, bindingsHeuristic);

			// The set of tasks which is always visited from the start, so there is no need for long distance dependencies
			Set<XEventClass> alwaysVisitedTasks = new HashSet<>();
			for (XEventClass task : eventRelations.getEventClasses()) {
				if (!existsPathWithoutTarget(eventRelations.getStartEventClass(), task, bindings)) {
					alwaysVisitedTasks.add(task);
				}
			}

			//TODO this might be wrong in lots of cases, what if the source node was activated in an AND?
			//	   theoretically, it needs to properly replay the whole state space to see whether the edge is needed! 

			Set<Relation> longTermDependencies = new HashSet<>();
			for (Relation relation : eventRelations.allNonEqualCombinations()) { // a!=b
				// long term dependency heuristic does not make sense from start or to end
				if (!eventRelations.includesStartOrEnd(relation)) {
					// not yet in our causal graph
					if (!causalGraph.getCausalRelations().contains(relation)) {
						DependencyRelation longDistanceDependency = eventRelations.getLongDistanceDependency(relation);
						if (longDistanceDependency.getDependencyValue() >= config.getLongTermDependencyThreshold()
								&& !alwaysVisitedTasks.contains(relation.getTarget())) {
							if (existsPathWithoutTarget(relation.getSource(), relation.getTarget(), bindings)) {
								longTermDependencies.add(relation);
							}
						}
					}
				}
			}

			if (!longTermDependencies.isEmpty()) {
				causalGraph = new SimpleCausalGraphImpl(causalGraph.getSetActivities(),
						Sets.union(causalGraph.getCausalRelations(), longTermDependencies));
			}

			if (config.isNoTasksUnconnected()) {
				causalGraph = removeUnconnected(causalGraph);
			}

			bindings = calculateBindings(context, causalGraph, bindingsHeuristic);
			CausalNet cnet = convertToCausalNet(causalGraph, bindings);

			return new FrequencyAwareCausalNetImpl(cnet, bindings, eventRelations.getEventClassesById());
		} else {

			if (config.isNoTasksUnconnected()) {
				causalGraph = removeUnconnected(causalGraph);
			}

			Bindings bindings = calculateBindings(context, causalGraph, bindingsHeuristic);
			CausalNet cnet = convertToCausalNet(causalGraph, bindings);

			return new FrequencyAwareCausalNetImpl(cnet, bindings, eventRelations.getEventClassesById());
		}
	}

	protected Bindings calculateBindings(final MinerContext context, SimpleCausalGraph causalGraph,
			BindingsHeuristic bindingsHeuristic) throws MiningException {
		//TODO I could keep this instance around and just use the updated threshold
		bindingsHeuristic.computeBindings(context, causalGraph);
		BindingsHeuristicConfig config = new BindingsHeuristicConfig();
		config.setBindingsThreshold(getConfig().getBindingsThreshold());
		return bindingsHeuristic.getFilteredBindings(config);
	}

	private SimpleCausalGraph removeUnconnected(final SimpleCausalGraph causalGraph) {
		Set<Relation> graph = new HashSet<>();
		for (Iterator<Relation> iterator = causalGraph.getCausalRelations().iterator(); iterator.hasNext();) {
			Relation relation = iterator.next();
			if (GraphUtils.isConnected(causalGraph.getCausalRelations(), relation, eventRelations.getStartEventClass(),
					eventRelations.getEndEventClass())) {
				graph.add(relation);
			}
		}
		return new SimpleCausalGraphImpl(causalGraph.getSetActivities(), graph);
	}

	private CausalNet convertToCausalNet(SimpleCausalGraph causalGraph, Bindings bindingInfo) {
		return CausalToCNet.convert(causalGraph, bindingInfo, eventRelations.getStartEventClass(),
				eventRelations.getEndEventClass()).getCausalNet();
	}

	private SimpleCausalGraph addStartEndActivities(SimpleCausalGraph causalGraph, XEventClass startActivity,
			XEventClass endActivity) {
		ImmutableSet<XEventClass> extendedActivitySet = ImmutableSet.<XEventClass>builder()
				.addAll(causalGraph.getSetActivities()).add(startActivity).add(endActivity).build();
		Builder<Relation> extendedRelations = ImmutableSet.<Relation>builder().addAll(causalGraph.getCausalRelations());
		for (XEventClass activity : causalGraph.getSetActivities()) {
			if (causalGraph.getPreset(activity).isEmpty()) {
				extendedRelations.add(Relation.Factory.create(startActivity, activity));
			}
			if (causalGraph.getPostset(activity).isEmpty()) {
				extendedRelations.add(Relation.Factory.create(activity, endActivity));
			}
		}
		return new SimpleCausalGraphImpl(extendedActivitySet, extendedRelations.build());
	}

	private void checkHasSourcesAndSinks(SimpleCausalGraph causalGraph) {
		boolean hasSource = false;
		boolean hasSink = false;
		for (XEventClass activity : causalGraph.getSetActivities()) {
			if (causalGraph.getPreset(activity).isEmpty()) {
				hasSource = true;
			}
			if (causalGraph.getPostset(activity).isEmpty()) {
				hasSink = true;
			}
		}
		if (!hasSink) {
			throw new IllegalArgumentException("Invalid causal graph, no sink! Please check your parameter settings!");
		}
		if (!hasSource) {
			throw new IllegalArgumentException(
					"Invalid causal graph, no source! Please check your parameter settings!");
		}
	}

	private void checkSameEventClasses(SimpleCausalGraph causalGraph) {
		Set<XEventClass> activities = causalGraph.getSetActivities();
		for (XEventClass eventClass : eventRelations.getEventClasses()) {
			if (!activities.contains(eventClass)) {
				throw new IllegalArgumentException(MessageFormat.format(
						"Causal graph is defined for a different set of activities: {0} is missing!", eventClass));
			}
		}
	}

	public CausalNetConfig getConfig() {
		return config;
	}

	public void setConfig(CausalNetConfig config) {
		this.config = config;
	}

	private boolean existsPathWithoutTarget(XEventClass source, XEventClass target, Bindings bindingInfo) {
		if (target.equals(eventRelations.getEndEventClass()) || source.equals(target)) { // every path includes the end node and there is always a path toward ourselves
			return false;
		} else {
			return recursivePathWithoutTarget(source, target, bindingInfo, new HashSet<XEventClass>());
		}
	}

	private boolean recursivePathWithoutTarget(XEventClass source, XEventClass target, Bindings bindings,
			Set<XEventClass> visitedNodes) {

		if (visitedNodes.contains(source) || source.equals(target)) {
			return false; // found the target / or loop
		} else if (source == eventRelations.getEndEventClass()) {
			return true; // found the end
		}

		for (Set<XEventClass> output : bindings.getOutputSets(source).elementSet()) {
			boolean pathWithoutTarget = true;
			for (XEventClass newSource : output) {
				ImmutableSet<XEventClass> recursiveVisitedNodes = ImmutableSet.<XEventClass>builder()
						.addAll(visitedNodes).add(source).build();
				if (!recursivePathWithoutTarget(newSource, target, bindings, recursiveVisitedNodes)) {
					pathWithoutTarget = false; // found output task in AND that has path to target 
				}
			}
			if (pathWithoutTarget) {
				return true; // found output set in XOR that has path without target
			}
		}

		return false; // none of the output sets in XOR had a path without target
	}

}