package org.processmining.dataawarecnetminer.mining.data;

import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;
import java.util.concurrent.Callable;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;

import org.apache.commons.lang3.StringEscapeUtils;
import org.apache.commons.math3.stat.StatUtils;
import org.deckfour.xes.classification.XEventClass;
import org.processmining.dataawarecnetminer.common.MinerContext;
import org.processmining.dataawarecnetminer.common.ProgressNoOpImpl;
import org.processmining.dataawarecnetminer.extension.conditionaldependencies.ConditionalDependencyHeuristicConfig;
import org.processmining.dataawarecnetminer.mining.classic.AcceptedConnectedHeuristic;
import org.processmining.dataawarecnetminer.mining.classic.AllTaskConnectedHeuristic;
import org.processmining.dataawarecnetminer.model.DataAwareCausalGraph;
import org.processmining.dataawarecnetminer.model.DataAwareCausalGraphImpl;
import org.processmining.dataawarecnetminer.model.DataRelation;
import org.processmining.dataawarecnetminer.model.DataRelationImpl;
import org.processmining.dataawarecnetminer.model.DataRelationStorage;
import org.processmining.dataawarecnetminer.model.DependencyAwareCausalGraph;
import org.processmining.dataawarecnetminer.model.DependencyRelation;
import org.processmining.dataawarecnetminer.model.DependencyRelationBinaryImpl;
import org.processmining.dataawarecnetminer.model.EventRelationStorage;
import org.processmining.dataawarecnetminer.util.DataDiscoveryUtil;
import org.processmining.datadiscovery.RuleDiscoveryException;
import org.processmining.datadiscovery.estimators.FunctionEstimation;
import org.processmining.datadiscovery.estimators.FunctionEstimator;
import org.processmining.datadiscovery.estimators.Type;
import org.processmining.datadiscovery.estimators.impl.DecisionTreeFunctionEstimator;
import org.processmining.datapetrinets.expression.GuardExpression;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.Progress;
import org.processmining.models.causalgraph.Relation;
import org.processmining.models.causalgraph.SimpleCausalGraph;

import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Multiset;
import com.google.common.collect.Multiset.Entry;

public class DataAwareCausalGraphBuilder {

	private final static class DecisionProblem {

		private final Relation relation;
		private final Set<Relation> otherRelations;

		public DecisionProblem(Relation relation, Set<Relation> otherRelations) {
			this.relation = relation;
			this.otherRelations = otherRelations;
		}

		public Set<Relation> getOtherRelations() {
			return otherRelations;
		}

		public Relation getPrimaryRelation() {
			return relation;
		}

		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + ((otherRelations == null) ? 0 : otherRelations.hashCode());
			result = prime * result + ((relation == null) ? 0 : relation.hashCode());
			return result;
		}

		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (getClass() != obj.getClass())
				return false;
			DecisionProblem other = (DecisionProblem) obj;
			if (otherRelations == null) {
				if (other.otherRelations != null)
					return false;
			} else if (!otherRelations.equals(other.otherRelations))
				return false;
			if (relation == null) {
				if (other.relation != null)
					return false;
			} else if (!relation.equals(other.relation))
				return false;
			return true;
		}

	}

	private final class DataDependencyTask implements Callable<Void> {

		private final Relation relation;
		private final Set<Relation> otherRelations;

		private final ConcurrentMap<Relation, DataRelation> dataRelations;
		private final ConcurrentMap<Relation, DependencyRelation> dependencyRelations;

		public DataDependencyTask(Relation relation, Set<Relation> otherRelations,
				ConcurrentMap<Relation, DataRelation> dataRelations,
				ConcurrentMap<Relation, DependencyRelation> dependencyRelations) {
			this.relation = relation;
			this.otherRelations = otherRelations;
			this.dataRelations = dataRelations;
			this.dependencyRelations = dependencyRelations;
		}

		public Void call() throws Exception {

			DataRelation dataDependency = getDataDependency();

			if (!dataDependency.isConditionTrue()) {

				if (dataDependency.getQualityScore() >= getConfig().getGuardThreshold()) {

					Iterable<Relation> candidates = ImmutableList.of(relation);

					for (Relation candidateRelation : candidates) {

						int requiredObservations = calcRequiredObservations(candidateRelation, dataDependency);

						// Only check relation that have not yet been accepted & are observed
						if (!dependencyRelations.containsKey(candidateRelation)) {

							boolean isFollowsRelationObserved = eventStore.countDirectlyFollows(candidateRelation) > 0;
							boolean isLoop = candidateRelation.getSource().equals(candidateRelation.getTarget());
							if (isFollowsRelationObserved && isLoop) {
								addPossibleL1Dependency(dataDependency, candidateRelation, requiredObservations);
							} else if (isFollowsRelationObserved) {
								addPossibleNormalDependency(dataDependency, candidateRelation, requiredObservations);
							}
						} else {

							if (DEBUG) {
								if (!(eventStore.countDirectlyFollows(candidateRelation) == 0)) {
									if (DEBUG) {
										System.out.println(MessageFormat.format(
												"\"{0}\",skipAlreadyPresent,\"{1}\",\"{2}\",{3,number,#},\"{4}\"",
												candidateRelation, dataDependency.getQualityScore(), 0.0, 0.0,
												escapedGuard(dataDependency.getCondition())));
									}
								}
							}

						}
					}
				} else {
					if (DEBUG) {
						System.out.println(
								MessageFormat.format("\"{0}\",skipBelowQuality,\"{1}\",\"{2}\",{3,number,#},\"{4}\"",
										relation, dataDependency.getQualityScore(), 0.0, 0.0,
										escapedGuard(dataDependency.getCondition())));
					}
				}

			}
			return null;
		}

		private DataRelation getDataDependency() throws RuleDiscoveryException {
			if (dataStore.getConsideredAttributes().isEmpty()) {
				return new DataRelationImpl(GuardExpression.Factory.trueInstance(), relation);
			} else {
				return computeDataRelation(new DecisionProblem(relation, otherRelations));
			}
		}

		private int calcRequiredObservations(Relation relation, DataRelation dataDependency) {
			if (getConfig().getObservationThreshold() == 0.0d) {
				return 1;
			}
			return 1;
		}

		private void addPossibleL1Dependency(DataRelation dataDependency, Relation candidateRelation,
				int requiredObservations) {
			// L1 loop
			GuardExpression condition = dataDependency.getCondition();
			DependencyRelation l1Relation = dataStore.getL1RelationUnderCondition(candidateRelation, condition);

			double dependencyValue = l1Relation.getDependencyValue();
			if (dependencyValue >= getConfig().getDependencyThreshold()) {
				int followsUnderCondition = dataStore.countDirectlyFollows(candidateRelation, condition);
				if (followsUnderCondition >= requiredObservations) {
					dataRelations.put(relation, dataDependency);
					dependencyRelations.put(candidateRelation, l1Relation);
					if (DEBUG) {
						System.out.println(MessageFormat.format("\"{0}\",added,\"{1}\",\"{2}\",{3,number,#},\"{4}\"",
								relation, dataDependency.getQualityScore(), dependencyValue, followsUnderCondition,
								escapedGuard(condition)));
					}
				} else {
					if (DEBUG) {
						System.out.println(
								MessageFormat.format("\"{0}\",skipObservation,\"{1}\",\"{2}\",{3,number,#},\"{4}\"",
										relation, dataDependency.getQualityScore(), dependencyValue,
										followsUnderCondition, escapedGuard(condition)));
					}
				}
			} else {
				if (DEBUG) {
					System.out.println(MessageFormat.format(
							"\"{0}\",skipDependency,\"{1}\",\"{2}\",{3,number,#},\"{4}\"", relation,
							dataDependency.getQualityScore(), dependencyValue,
							dataStore.countDirectlyFollows(candidateRelation, condition), escapedGuard(condition)));
				}
			}
		}

		private void addPossibleNormalDependency(DataRelation dataDependency, Relation candidateRelation,
				int requiredObservations) {

			GuardExpression condition = dataDependency.getCondition();
			DependencyRelation normalRelation = dataStore.getDependencyRelationUnderCondition(candidateRelation,
					condition);

			double dependencyValue = normalRelation.getDependencyValue();
			if (dependencyValue >= getConfig().getDependencyThreshold()) {
				int followsUnderCondition = dataStore.countDirectlyFollows(candidateRelation, condition);
				if (followsUnderCondition >= requiredObservations) {
					dataRelations.put(relation, dataDependency);
					dependencyRelations.put(candidateRelation, normalRelation);
					if (DEBUG) {
						System.out.println(MessageFormat.format("\"{0}\",added,\"{1}\",\"{2}\",{3,number,#},\"{4}\"",
								relation, dataDependency.getQualityScore(), dependencyValue, followsUnderCondition,
								escapedGuard(condition)));
					}
				} else {
					if (DEBUG) {
						System.out.println(
								MessageFormat.format("\"{0}\",skipObservation,\"{1}\",\"{2}\",{3,number,#},\"{4}\"",
										relation, dataDependency.getQualityScore(), dependencyValue,
										followsUnderCondition, escapedGuard(condition)));
					}
				}
			} else {
				if (DEBUG) {
					System.out.println(MessageFormat.format(
							"\"{0}\",skipDependency,\"{1}\",\"{2}\",{3,number,#},\"{4}\"", relation,
							dataDependency.getQualityScore(), dependencyValue,
							dataStore.countDirectlyFollows(candidateRelation, condition), escapedGuard(condition)));
				}
			}
		}

		private String escapedGuard(GuardExpression condition) {
			return StringEscapeUtils.escapeCsv(condition.toCanonicalString());
		}

		public DataRelation computeDataRelation(DecisionProblem decisionProblem) throws RuleDiscoveryException {

			Relation relation = decisionProblem.getPrimaryRelation();
			Set<Relation> otherRelations = decisionProblem.getOtherRelations();

			String thisClass = relation.getTarget().getId();
			String otherClass = "NOT " + relation.getTarget().getId();

			DecisionTreeFunctionEstimator estimator = createEstimator(
					DataDiscoveryUtil.viewSelectedOnly(dataStore.getAttributeTypes(),
							dataStore.getConsideredAttributes()),
					dataStore.getLiteralValues(), relation.toString(), new Object[] { thisClass, otherClass });

			estimator.setBinarySplit(getConfig().getDataDiscoveryConfig().isBinarySplit());
			estimator.setConfidenceFactor(getConfig().getDataDiscoveryConfig().getConfidenceThreshold());
			estimator.setCrossValidate(getConfig().getDataDiscoveryConfig().isCrossValidate());
			estimator.setUnpruned(getConfig().getDataDiscoveryConfig().isUnpruned());
			estimator.setTreatNoLeafAsFalse(false);
			estimator.setCrossValidateRandom(new Random(1486608900));

			addInstancesForRelation(thisClass, relation, estimator, dataStore.getConsideredAttributes());
			for (Relation otherRelation : otherRelations) {
				addInstancesForRelation(otherClass, otherRelation, estimator, dataStore.getConsideredAttributes());
			}

			int numThisInstances = eventStore.countDirectlyFollows(relation);
			if (numThisInstances > 1) {
				estimator.setNumFoldCrossValidation(Math.min(config.getNumCrossValidations(), numThisInstances));
				double occurenceCount = estimator.getSumOfWeights();
				int minNumObj = (int) Math.max(2, Math
						.ceil(occurenceCount * getConfig().getDataDiscoveryConfig().getMinPercentageObjectsOnLeaf()));
				estimator.setMinNumObj(minNumObj);

				FunctionEstimation estimation = null;
				int repetitions = estimator.getNumFoldCrossValidation();
				double[] qualities = new double[repetitions];
				for (int i = 0; i < repetitions; i++) {
					estimation = getEstimation(estimator, thisClass);
					switch (config.getQualityMeasure()) {
						case FSCORE :
							qualities[i] = estimator.getEvaluation().weightedFMeasure();
							break;
						case KAPPA :
						default :
							qualities[i] = estimator.getEvaluation().kappa();
							break;
					}

				}
				double qualityMean = StatUtils.mean(qualities);
				double qualityVariance = StatUtils.variance(qualities);

				return new DataRelationImpl(estimation.getExpression(), relation, qualityMean, qualityVariance,
						estimator.getClassifier(), estimator.getEvaluation());
			} else {
				return new DataRelationImpl(GuardExpression.Factory.trueInstance(), relation);
			}

		}

		private FunctionEstimation getEstimation(DecisionTreeFunctionEstimator estimator, String thisClass)
				throws RuleDiscoveryException {
			try {
				Map<Object, FunctionEstimation> result = estimator.getFunctionEstimation(null);
				FunctionEstimation dataDependency = result.get(thisClass);
				return dataDependency;
			} catch (Exception e) {
				throw new RuleDiscoveryException("Error creating decision tree", e);
			}
		}

		private void addInstancesForRelation(String clazz, Relation relation, FunctionEstimator estimator,
				Set<String> consideredAttributes) {

			Multiset<Map<String, Object>> observedAttributes = dataStore.directlyFollowsUnderAttributes(relation);
			for (Entry<Map<String, Object>> instanceEntry : observedAttributes.entrySet()) {
				addInstance(clazz, estimator, instanceEntry.getElement(), instanceEntry.getCount());
			}

		}

		private void addInstance(String clazz, FunctionEstimator estimator, Map<String, Object> attributes, int count) {
			try {
				estimator.addInstance(attributes, clazz, count);
			} catch (Exception e) {
				throw new RuntimeException("Could not add instance", e);
			}
		}

	}

	private static final boolean DEBUG = true;

	private final EventRelationStorage eventStore;
	private final DataRelationStorage dataStore;

	private ConditionalDependencyHeuristicConfig config;

	public DataAwareCausalGraphBuilder(EventRelationStorage relations, DataRelationStorage dataRelations) {
		this(relations, dataRelations, new ConditionalDependencyHeuristicConfig());
	}

	public DataAwareCausalGraphBuilder(EventRelationStorage relations, DataRelationStorage dataRelations,
			ConditionalDependencyHeuristicConfig config) {
		this.eventStore = relations;
		this.dataStore = dataRelations;
		this.config = config;
	}

	public DataAwareCausalGraph build(SimpleCausalGraph initialCausalGraph) throws RuleDiscoveryException {
		final ExecutorService executor = Executors.newFixedThreadPool(Runtime.getRuntime().availableProcessors());
		try {
			return build(new MinerContext() {

				public Progress getProgress() {
					return new ProgressNoOpImpl();
				}

				public ExecutorService getExecutor() {
					return executor;
				}

				public PluginContext getPluginContext() {
					return null;
				}
			}, initialCausalGraph);
		} finally {
			executor.shutdown();
		}
	}

	public DataAwareCausalGraph build(MinerContext context, SimpleCausalGraph causalGraph)
			throws RuleDiscoveryException {

		if (DEBUG) {
			System.out.println("Discovering data dependencies using attributes: ");
			System.out.println(config.getDataDiscoveryAttributes());
			System.out.println("Parameters:");
			System.out.println("DEP: " + config.getDependencyThreshold());
			System.out.println("OBS: " + config.getObservationThreshold());
			System.out.println("CON: " + config.getGuardThreshold());
			System.out.println("relation,action,conditionQuality,dependencyScore,numObservations,rule");
		}

		ConcurrentMap<Relation, DependencyRelation> dependencyRelations = convertOriginalRelations(causalGraph);

		if (config.getDataDiscoveryAttributes().isEmpty()) {
			postProcess(dependencyRelations);
			return new DataAwareCausalGraphImpl(eventStore.getEventClasses(), dependencyRelations,
					ImmutableMap.<Relation, DataRelation>of());
		} else {
			dataStore.setConsideredAttributes(config.getDataDiscoveryAttributes());
			ConcurrentMap<Relation, DataRelation> dataRelations = computeDataDependencies(context.getExecutor(),
					dependencyRelations);
			postProcess(dependencyRelations);

			if (DEBUG) {
				System.out.println("#Relations: " + dependencyRelations.size());
				System.out.println("#Data-Relations: " + dataRelations.size());
			}
			return new DataAwareCausalGraphImpl(eventStore.getEventClasses(), dependencyRelations, dataRelations);
		}

	}

	private void postProcess(ConcurrentMap<Relation, DependencyRelation> dependencyRelations) {
		if (config.isAcceptedTasksConnected()) {
			AcceptedConnectedHeuristic.connectAcceptedTasks(eventStore, dependencyRelations,
					getConfig().getRelativeToBestThreshold());
		}
		if (config.isAllTasksConnected()) {
			AllTaskConnectedHeuristic.connectAllTasks(eventStore, dependencyRelations,
					getConfig().getRelativeToBestThreshold());
		}
	}

	private ConcurrentMap<Relation, DependencyRelation> convertOriginalRelations(SimpleCausalGraph causalGraph) {
		ConcurrentMap<Relation, DependencyRelation> dependencyRelations = new ConcurrentHashMap<>();
		if (causalGraph instanceof DependencyAwareCausalGraph) {
			DependencyAwareCausalGraph dependencyAwareCausalGraph = (DependencyAwareCausalGraph) causalGraph;
			for (Relation relation : causalGraph.getCausalRelations()) {
				dependencyRelations.put(relation, dependencyAwareCausalGraph.getDependencyRelation(relation));
			}
		} else {
			DependencyRelation binaryRelation = new DependencyRelationBinaryImpl();
			for (Relation relation : causalGraph.getCausalRelations()) {
				dependencyRelations.put(relation, binaryRelation);
			}
		}
		return dependencyRelations;
	}

	private ConcurrentMap<Relation, DataRelation> computeDataDependencies(ExecutorService pool,
			final ConcurrentMap<Relation, DependencyRelation> dependencyRelations) throws RuleDiscoveryException {

		ConcurrentMap<Relation, DataRelation> dataRelations = new ConcurrentHashMap<>();

		List<Future<Void>> results = new ArrayList<>(eventStore.getEventClasses().size() * 4);
		for (Relation relation : eventStore.allObservedCombinations()) {
			if (!dependencyRelations.containsKey(relation)) {
				Set<Relation> otherRelations = findOthers(relation);
				if (!otherRelations.isEmpty()) {
					// only consider relations which are not yet found and observed
					results.add(pool.submit(
							new DataDependencyTask(relation, otherRelations, dataRelations, dependencyRelations)));
				}
			}
		}

		try {
			for (Future<Void> result : results) {
				result.get();
			}
		} catch (InterruptedException e) {
			Thread.currentThread().interrupt();
		} catch (ExecutionException e) {
			throw new RuleDiscoveryException(e);
		}

		return dataRelations;
	}

	protected Set<Relation> findOthers(Relation relation) {
		// get all observed directly following classes
		Collection<XEventClass> otherTargets = eventStore.directlyFollows(relation.getSource());
		Set<Relation> others = new LinkedHashSet<>(otherTargets.size());
		double dependencyThreshold = getConfig().getDependencyThreshold();
		for (XEventClass otherTarget : otherTargets) {
			//	Only observed relations other than (a,b) itself
			if (!otherTarget.equals(relation.getTarget())) {
				Relation otherRelation = Relation.Factory.create(relation.getSource(), otherTarget);
				DependencyRelation dependencyRelation = eventStore.getDependencyRelation(otherRelation);
				// Only relations (a,s), which are true dependencies according to the threshold				
				if (dependencyRelation.getDependencyValue() >= dependencyThreshold) {
					others.add(otherRelation);
				}
			}
		}
		return others;
	}

	protected DecisionTreeFunctionEstimator createEstimator(Map<String, Type> attributeType,
			Map<String, Set<String>> literalValues, String edgeDesc, Object[] classes) {
		return new DecisionTreeFunctionEstimator(attributeType, literalValues, classes, edgeDesc, 1024);
	}

	public ConditionalDependencyHeuristicConfig getConfig() {
		return config;
	}

	public void setConfig(ConditionalDependencyHeuristicConfig config) {
		this.config = config;
	}

}