package org.processmining.dataawarecnetminer.mining.decision;

import java.util.Collection;
import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.concurrent.Callable;

import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XTrace;
import org.processmining.dataawarecnetminer.extension.decision.DecisionHeuristicConfig;
import org.processmining.datadiscovery.ProjectedLog;
import org.processmining.datadiscovery.RuleDiscovery;
import org.processmining.datadiscovery.RuleDiscovery.Rule;
import org.processmining.datadiscovery.WekaDecisionTreeRuleDiscovery.TreeRule;
import org.processmining.datadiscovery.estimators.FunctionEstimation;
import org.processmining.datapetrinets.expression.GuardExpression;
import org.processmining.models.cnet.CNetBinding;
import org.processmining.models.cnet.CNetNode;

import com.google.common.collect.ImmutableSet;

import weka.classifiers.Evaluation;

abstract class DefaultGuardedBindingDiscovery implements Callable<Set<GuardedBinding>> {

	private final DecisionHeuristicConfig config;

	private final Collection<XTrace> traces;
	private final XEvent artificalStart;
	private final XEvent artificalEnd;

	private final ActivityMapper activityMapper;

	private final CNetNode node;
	private final Set<CNetBinding> bindings;

	DefaultGuardedBindingDiscovery(Collection<XTrace> traces, CNetNode node, Set<CNetBinding> bindings,
			XEvent artificalStart, XEvent artificalEnd, ActivityMapper activityMapper, DecisionHeuristicConfig config) {
		this.traces = traces;
		this.activityMapper = activityMapper;
		this.node = node;
		this.bindings = bindings;
		this.artificalStart = artificalStart;
		this.artificalEnd = artificalEnd;
		this.config = config;
	}

	abstract protected RuleDiscovery<? extends Rule> getRuleDiscoveryMethod(ProjectedLog projectedLog);

	public Set<GuardedBinding> call() throws Exception {
		Set<GuardedBinding> guardedBindings = new HashSet<>();

		ProjectedLog projectedLog = createObservationInstances(activityMapper);
		RuleDiscovery<? extends Rule> ruleDiscovery = getRuleDiscoveryMethod(projectedLog);

		CNetBinding[] outputClasses = bindings.toArray(new CNetBinding[bindings.size()]);

		final Rule rule = ruleDiscovery.discover(node, ImmutableSet.copyOf(outputClasses));

		Map<Object, FunctionEstimation> estimation = rule.getRules();

		// Just use all discovered rules
		for (final Entry<Object, FunctionEstimation> entry : estimation.entrySet()) {
			if (!config.usesDecisionThreshold() || entry.getValue().getQualityMeasure() >= config.getDecisionThreshold())
				guardedBindings.add(new GuardedBinding() {

					public FunctionEstimation getFunctionEstimation() {
						return new FunctionEstimation() {

							public Double getQualityMeasure() {
								return entry.getValue().getQualityMeasure();
							}

							public GuardExpression getExpression() {
								return entry.getValue().getExpression();
							}

							public String toString() {
								if (rule instanceof TreeRule) {
									Evaluation evaluation = ((TreeRule) rule).getEvaluation();
									try {
										return evaluation.toMatrixString() + "\n" + evaluation.toSummaryString(false)
												+ "\n" + evaluation.toClassDetailsString();
									} catch (Exception e) {
									}
								}
								return "Quality: " + entry.getValue().getQualityMeasure();
							}

						};
					}

					public CNetBinding getBinding() {
						return (CNetBinding) entry.getKey();
					}

				});
		}

		return guardedBindings;
	}

	private ProjectedLog createObservationInstances(ActivityMapper activityMapper) {
		return new ProjectedLogImpl(traces, activityMapper, artificalStart, artificalEnd,
				config.getDataDiscoveryAttributes());
	}

}