package org.processmining.dataawarecnetminer.model;

import java.util.HashSet;
import java.util.Set;

import org.deckfour.xes.classification.XEventClass;
import org.processmining.dataawarecnetminer.extension.bindings.Bindings;
import org.processmining.models.cnet.CNetBinding;
import org.processmining.models.cnet.CNetBinding.Type;
import org.processmining.models.cnet.CNetNode;
import org.processmining.models.cnet.CausalNet;

import com.google.common.collect.ImmutableMap;
import com.google.common.collect.Multiset;

public class FrequencyAwareCausalNetImpl implements FrequencyAwareCausalNet {

	private final CausalNet cnet;
	private final Bindings bindings;
	private final ImmutableMap<String, XEventClass> eventClasses;

	public FrequencyAwareCausalNetImpl(CausalNet cnet, Bindings bindings,
			ImmutableMap<String, XEventClass> eventClasses) {
		this.cnet = cnet;
		this.bindings = bindings;
		this.eventClasses = eventClasses;
	}

	public CausalNet getCNet() {
		return cnet;
	}

	public int getBindingFrequency(CNetBinding binding) {
		XEventClass eventClass = getEventClassForNode(binding.getNode());
		if (binding.getType() == Type.INPUT) {
			return getBindingFrequency(binding, bindings.getInputSets(eventClass));
		} else {
			return getBindingFrequency(binding, bindings.getOutputSets(eventClass));
		}
	}

	private int getBindingFrequency(CNetBinding binding, Multiset<Set<XEventClass>> nodeBindings) {
		Set<XEventClass> translatedBinding = new HashSet<>();
		for (CNetNode node : binding.getBoundNodes()) {
			translatedBinding.add(getEventClassForNode(node));
		}
		return nodeBindings.count(translatedBinding);
	}

	private XEventClass getEventClassForNode(CNetNode node) {
		return eventClasses.get(node.getLabel());
	}

	public String toString() {
		return cnet.toString();
	}

	public int getInputFrequency(CNetNode node) {
		XEventClass eventClass = getEventClassForNode(node);
		Multiset<Set<XEventClass>> inputSets = bindings.getInputSets(eventClass);
		if (inputSets != null) {
			return inputSets.size();	
		} else {
			return 0;
		}
	}

	public int getOutputFrequency(CNetNode node) {
		XEventClass eventClass = getEventClassForNode(node);
		Multiset<Set<XEventClass>> outputSets = bindings.getOutputSets(eventClass);
		if (outputSets != null) {
			return outputSets.size();
		} else {
			return 0;
		}
	}

	public int getNodeFrequency(CNetNode node) {
		return Math.max(getInputFrequency(node), getOutputFrequency(node));
	}

	public int getNodeSkippedFrequency(CNetNode node) {
		return 0;
	}

	public int getNodeInsertedFrequency(CNetNode node) {
		return 0;
	}

	public int getCorrectVariableFrequency(CNetNode node, String variable) {
		return 0;
	}

	public int getMissingVariableFrequency(CNetNode node, String variable) {
		return 0;
	}

	public int getIncorrectVariableFrequency(CNetNode node, String variable) {
		return 0;
	}
}