package org.processmining.dataawarecnetminer.visualization;

import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Set;

import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.DefaultListModel;
import javax.swing.JCheckBox;
import javax.swing.JPanel;
import javax.swing.JSlider;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

import org.processmining.dataawarecnetminer.interactive.InteractiveDataAwareCausalMinerVisualization;
import org.processmining.dataawarecnetminer.util.ExtendedDoubleSlider;
import org.processmining.framework.util.ui.widgets.ProMList;

import com.fluxicon.slickerbox.components.NiceSlider.Orientation;
import com.fluxicon.slickerbox.factory.SlickerFactory;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Ordering;

public class DataDiscoveryConfigurationPanel extends JPanel {

	private static final long serialVersionUID = 1L;

	private final ProMList<String> attributesSelection;
	private final DefaultListModel<String> attributesSelectionListModel;

	private final ExtendedDoubleSlider minLeafSlider;
	private final JCheckBox crossValidateBox;
	private final JCheckBox binarySplitBox;

	public DataDiscoveryConfigurationPanel(Set<String> availableAttributes) {
		this(availableAttributes, null);
	}

	public DataDiscoveryConfigurationPanel(Set<String> availableAttributes,
			final InteractiveDataAwareCausalMinerVisualization interactiveVisualization) {
		super();

		setLayout(new BorderLayout());
		attributesSelectionListModel = new DefaultListModel<>();
		for (String c : Ordering.natural().immutableSortedCopy(availableAttributes)) {
			attributesSelectionListModel.addElement(c);
		}
		attributesSelection = new ProMList<>("Considered attributes", attributesSelectionListModel);
		attributesSelection.setPreferredSize(null);
		attributesSelection.getList().addListSelectionListener(new ListSelectionListener() {

			public void valueChanged(ListSelectionEvent e) {
				if (interactiveVisualization != null) {
					// interactive mode
					interactiveVisualization.update();
				}
			}

		});
		add(attributesSelection, BorderLayout.CENTER);

		JPanel bottomPanel = new JPanel();
		bottomPanel.setLayout(new BoxLayout(bottomPanel, BoxLayout.Y_AXIS));
		bottomPanel.setOpaque(false);

		JPanel sliderPanel = new JPanel();
		sliderPanel.setLayout(new BoxLayout(sliderPanel, BoxLayout.Y_AXIS));
		sliderPanel.setOpaque(false);

		minLeafSlider = new ExtendedDoubleSlider("Minimum instances used by C4.5 (relative)", 0.0, 0.5, 0.01,
				Orientation.HORIZONTAL);
		minLeafSlider.getSlider().addChangeListener(new ChangeListener() {

			public void stateChanged(ChangeEvent e) {
				JSlider source = (JSlider) e.getSource();
				if (!source.getValueIsAdjusting()) {
					if (interactiveVisualization != null) {
						// interactive mode
						interactiveVisualization.update();
					}

				}
			}
		});
		sliderPanel.add(minLeafSlider);
		bottomPanel.add(sliderPanel);

		bottomPanel.add(Box.createHorizontalStrut(10));

		JPanel checkboxPanel = new JPanel();
		checkboxPanel.setLayout(new BoxLayout(checkboxPanel, BoxLayout.X_AXIS));
		checkboxPanel.setOpaque(false);

		crossValidateBox = SlickerFactory.instance()
				.createCheckBox("Cross validate classifier (10-fold, 10 repetitions)", false);
		crossValidateBox.addActionListener(new ActionListener() {

			public void actionPerformed(ActionEvent e) {
				if (interactiveVisualization != null) {
					// interactive mode
					interactiveVisualization.update();
				}
			}
		});
		checkboxPanel.add(crossValidateBox);
		binarySplitBox = SlickerFactory.instance().createCheckBox("Force binary rules", false);
		binarySplitBox.addActionListener(new ActionListener() {

			public void actionPerformed(ActionEvent e) {
				if (interactiveVisualization != null) {
					// interactive mode
					interactiveVisualization.update();
				}
			}
		});
		checkboxPanel.add(binarySplitBox);
		bottomPanel.add(checkboxPanel);

		add(bottomPanel, BorderLayout.SOUTH);
	}

	public void setSelectedAttributes(Set<String> dataDiscoveryAttributes) {
		int[] selectedIndicies = new int[dataDiscoveryAttributes.size()];
		int j = 0;
		for (int i = 0; i < attributesSelectionListModel.getSize(); i++) {
			String element = attributesSelectionListModel.getElementAt(i);
			if (dataDiscoveryAttributes.contains(element)) {
				selectedIndicies[j++] = i;
			}
		}
		attributesSelection.setSelectedIndices(selectedIndicies);
	}

	public void setMinPercentageObjectsOnLeaf(double minPercentageObjectsOnLeaf) {
		minLeafSlider.setValue(minPercentageObjectsOnLeaf);
	}

	public void setCrossValidate(boolean crossValidate) {
		crossValidateBox.setSelected(crossValidate);
	}

	public void setBinarySplit(boolean binarySplit) {
		binarySplitBox.setSelected(binarySplit);
	}

	public double getMinPercentageObjectsOnLeaf() {
		return minLeafSlider.getValue();
	}

	@SuppressWarnings("unchecked") // stupid ProMList
	public Set<String> getSelectedAttributes() {
		return ImmutableSet.copyOf(attributesSelection.getList().getSelectedValuesList());
	}

	public boolean isCrossCalidate() {
		return crossValidateBox.isSelected();
	}

	public boolean isBinarySplit() {
		return binarySplitBox.isSelected();
	}

}